<!DOCTYPE html>
<html lang='en'>
<head>
  <meta charset='utf-8'>
  <title>Sign in</title>
  <meta name='viewport' content='width=device-width, initial-scale=1.0'>
  <meta name='description' content=''>
  <meta name='author' content=''>
  <link href="//netdna.bootstrapcdn.com/twitter-bootstrap/2.2.2/css/bootstrap-combined.min.css" rel="stylesheet">
  <link rel="shortcut icon" href="//www.iconj.com/ico/w/d/wdgwj3n50t.ico" type="image/x-icon" />
  <style type="text/css">
        body {
          padding: 50px;
          background-color: #111;
        }
        .box {
          max-width: 220px;
          padding: 20px 50px;
          margin: 0 auto 20px;
          background-color: #fff;
          border: 1px solid #e5e5e5;
          -webkit-border-radius: 5px;
             -moz-border-radius: 5px;
                  border-radius: 5px;
          opacity: 0.8;
        }
        .box a {
          margin-bottom: 20px;
        }
  </style>
</head>
<body>
  <div class='box'>
      {% set messages = get_flashed_messages() %}
      {% if messages %}
        <h2>{{ messages[-1] }}</h2>
      {% else %}
        <h2>Welcome</h2>
      {% endif %}
    <a href='{{ sign_in_url }}' class='btn btn-primary btn-large btn-block'>Sign in with Google</a>
  </div>
</body>
</html>
